/*
  device - VPD support routines for devices.
  
  This file is part of the Linux lsvpd package.
  
  (C) Copyright IBM Corp. 2005
  
  Maintained by  Martin Schwenke <martins@au.ibm.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
 
  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
  $Id: device.c,v 1.6 2005/07/09 23:04:14 martins Exp $
*/

/* For asprintf(3). */
#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/param.h>
#include <sys/sysmacros.h>

#include "util.h"
#include "vpd.h"

#include "lsvpd.h"
#include "init.h"

#include "device.h"

/********************************************************************/

const char *
device_find_address(const device_t type,
		    const device_bus_info_t bus_info)
{
	const char * ret;
	size_t ldst;
	size_t lbi;

	ret = NULL;

	ldst = strlen(type);
	lbi = strlen(bus_info);

	if ((lbi > ldst) &&
	    (0 == memcmp(bus_info, type, ldst)) &&
	    (DEVICE_BUS_INFO_SEP == bus_info[ldst]))
		ret = bus_info + ldst + 1;

	return ret;
}

/********************************************************************/

device_functions_t * device_functions = NULL;

static device_functions_t *
device_find_type(const device_t type)
{
	device_functions_t * t;

	t = device_functions;

	while ((NULL != t) && (0 != strcmp(t->type, type))) {
		t = t->next;
	}

	return t;
}

#define DEVICE_SET_POINTER(f) \
		if (DEVICE_UNDEF == i->f) \
			d->f = NULL; \
		else if (NULL != i->f) \
			d->f = i->f;


/* Copies i if necessary. */
void
device_type_register(device_functions_t * i)
{
	device_functions_t * t;
	device_functions_t * d;

	t = device_find_type(i->type);

	if (NULL != t) {
		d = t;
	} else {
		d = malloc(sizeof(device_functions_t));
		memset(d, 0, sizeof(device_functions_t));
		d->type = i->type;
	}

	if (NULL != d) {
		DEVICE_SET_POINTER(supertype);
		DEVICE_SET_POINTER(validate);
		DEVICE_SET_POINTER(retrieve_vpd);
		DEVICE_SET_POINTER(get_parent_name);
		DEVICE_SET_POINTER(vpd_sort);
		DEVICE_SET_POINTER(get_info);
		DEVICE_SET_POINTER(get_kernel_name);
		DEVICE_SET_POINTER(get_name);
		//DEVICE_SET_POINTER(set_yl);
		//DEVICE_SET_POINTER(add_hook);

		if (NULL == t) {
			d->next = device_functions;
			device_functions = d;
		}
	}
}

#define DEVICE_MULTIPLEXED_VOID(name, argdecls, args) \
void \
device_ ## name argdecls \
{ \
	device_functions_t * t; \
	void (*f) argdecls; \
\
	t = device_find_type(type); \
	f = NULL; \
	while ((NULL == f) && (NULL != t)) { \
		f = t->name; \
		if (NULL == f) \
			t = (NULL != t->supertype) ? \
				device_find_type(t->supertype) : \
				NULL; \
	} \
\
	if (NULL != f) \
		(*f) args; \
}

#define DEVICE_MULTIPLEXED_RET(name, argdecls, args, ret_t, ret_def) \
ret_t \
device_ ## name argdecls \
{ \
	ret_t ret; \
	device_functions_t * t; \
	ret_t (*f) argdecls; \
\
	ret = ret_def; \
\
	t = device_find_type(type); \
	f = NULL; \
	while ((NULL == f) && (NULL != t)) { \
		f = t->name; \
		if (NULL == f) \
			t = (NULL != t->supertype) ? \
				device_find_type(t->supertype) : \
				NULL; \
	} \
\
	if (NULL != f) \
		ret = (*f) args; \
\
	return ret; \
}

DEVICE_MULTIPLEXED_RET(validate,
		       (const device_t type,
			const device_bus_info_t bus_info),
		       (type, bus_info),
		       bool,
		       true)

DEVICE_MULTIPLEXED_RET(retrieve_vpd,
		       (const device_t type,
			const device_bus_info_t bus_info,
			const char * tmp_dir,
			vpd_list_t ** vpd),
		       (type, bus_info, tmp_dir, vpd),
		       bool,
		       false)

DEVICE_MULTIPLEXED_RET(get_parent_name,
		       (const device_t type,
			const device_bus_info_t bus_info),
		       (type, bus_info),
		       char *,
		       NULL)

DEVICE_MULTIPLEXED_VOID(vpd_sort,
			(const device_t type,
			 vpd_list_t ** vpd),
			(type, vpd))

DEVICE_MULTIPLEXED_RET(get_info,
		       (const device_t type,
			const device_bus_info_t bus_info),
		       (type, bus_info),
		       device_node_info_t *,
		       NULL)

DEVICE_MULTIPLEXED_RET(get_name,
		       (const device_t type,
			const device_bus_info_t bus_info),
		       (type, bus_info),
		       char *,
		       NULL)


DEVICE_MULTIPLEXED_RET(get_kernel_name,
		       (const device_t type,
			const device_bus_info_t bus_info),
		       (type, bus_info),
		       char *,
		       NULL)

/********************************************************************/

/* Device listing. */

device_listing_function_list_t * device_listing_functions = NULL;

void device_listing_functions_clear()
{
	device_listing_function_list_t * t1;
	device_listing_function_list_t * t2;

	t1 = device_listing_functions;
	while (NULL != t1) {
		t2 = t1->next;
		free(t1);
		t1 = t2;
	}
	device_listing_functions = NULL;
}

void
device_listing_function_register(device_listing_function_t * f)
{
	device_listing_function_list_t * t;

	t = malloc(sizeof(device_listing_function_list_t));
	if (NULL != t) {
		t->f = f;
		t->next = device_listing_functions;
		device_listing_functions = t;
	}
}

bool
device_list_devices(device_bus_info_list_t ** device_list)
{
	bool ret;
	device_listing_function_list_t * t;

	ret = true;

	*device_list = NULL;
	for (t = device_listing_functions; NULL != t; t = t->next)
		ret = ret && t->f(device_list);

	return ret;
}

bool
device_list_add(device_bus_info_list_t ** device_list,
		const device_t type,
		const device_bus_info_t bus_info)
{
	bool ret;
	device_bus_info_list_t * t;

	ret = false;

	if (device_validate(type, bus_info)) {
		if (NULL != (t = malloc(sizeof(device_bus_info_list_t)))) {
			if (NULL != (t->bus_info = strdup(bus_info))) {
				t->next = *device_list;
				*device_list = t;
				ret = true;
			} else
				free(t);
		}
	} else
		/* This is an unreported "failure". */
		ret= true;

	return ret;
}

void
device_list_clear(device_bus_info_list_t ** device_list)
{
	device_bus_info_list_t * t1;
	device_bus_info_list_t * t2;

	t1 = *device_list;
	while (NULL != t1) {
		t2 = t1->next;
		free(t1->bus_info);
		free(t1);
		t1 = t2;
	}
	*device_list = NULL;
}

/********************************************************************/

void
device_close_temporary(device_temp_t ** d)
{
	if (NULL != *d) {
		if (-1 != (*d)->fd)
			close((*d)->fd);
		if (NULL != (*d)->name) {
			unlink((*d)->name);
			free((*d)->name);
		}
		free(*d);
	}
	*d = NULL;
}

extern device_temp_t *
device_open_temporary(const device_node_info_t * info)
{
	device_temp_t * ret;

	if (NULL != (ret = malloc(sizeof(device_temp_t)))) {
		ret->fd = -1;
		if (-1 != asprintf(&ret->name, "%s/node-%u-%u-%u",
				   lsvpd_db_temp_dir,
				   (unsigned int) info->mode,
				   info->major, info->minor))
			if (0 == mknod(ret->name, 
				       0660 | info->mode,
				       makedev(info->major, info->minor))) {
				ret->fd = open(ret->name, O_RDONLY|O_NONBLOCK);
				if (-1 == ret->fd)
					device_close_temporary(&ret);
			} else
				device_close_temporary(&ret);
		else
			device_close_temporary(&ret);
	}
	return ret;
}

/********************************************************************/

static const char * device_default_vpd_sort_order[] =
{
	"FC", "DS", "AX", "MF", "TM", "SN", "CD", "RL", "RM", "YL",
	NULL
};

static void
device_vpd_sort_default (const device_t type,
			 vpd_list_t ** vpd)
{
	vpd_list_sort(vpd, device_default_vpd_sort_order);
}

static char *
device_get_name_default(const device_t type,
			const device_bus_info_t bus_info)
{
	return device_get_kernel_name(type,
				      bus_info);
}

/********************************************************************/

/*
 * No need to define dummy functions for all functions, since those
 * functions will do the appropriate thing (fail) is no functions are
 * defined in device_functions.
 */

static void
init(void)
{
	device_functions_t i = 
		{
			.type     = DEVICE_DEFAULT_TYPE,
			.vpd_sort = device_vpd_sort_default,
			.get_name = device_get_name_default,
		};
	device_type_register(&i);
}

INIT(init);
