#ifndef LSVPD_DEVICE_H
#define LSVPD_DEVICE_H

/*
  lsvpd - Device-related definitions for the lsvpd package.
  
  This file is part of the Linux lsvpd package.
  
  (C) Copyright IBM Corp. 2002, 2003, 2004, 2005
  
  Maintained by  Martin Schwenke <martins@au.ibm.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
 
  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
  $Id: device.h,v 1.5 2005/07/09 23:05:09 martins Exp $
*/
  
#include <sys/types.h>
#include <stdbool.h>

#include "vpd.h"

typedef const char * device_t;
typedef char * device_bus_info_t;

#define DEVICE_BUS_INFO_SEP '/'

extern const char *
device_find_address(const device_t type,
		    const device_bus_info_t bus_info);

typedef struct {
	mode_t mode;
	unsigned int major;
	unsigned int minor;
} device_node_info_t;

/* For device listing.  This list is logically unordered. */
typedef struct device_bus_info_list {
	device_bus_info_t bus_info;
	struct device_bus_info_list * next;
} device_bus_info_list_t;

/********************************************************************/

typedef bool
(device_validate_function_t) (const device_t type,
			      const device_bus_info_t bus_info);

typedef bool
(device_retrieve_vpd_function_t) (const device_t type,
				  const device_bus_info_t bus_info,
				  const char * tmp_dir,
				  vpd_list_t ** vpd);

typedef char *
(device_get_parent_name_function_t) (const device_t type,
				     const device_bus_info_t bus_info);

typedef void
(device_vpd_sort_function_t) (const device_t type,
			      vpd_list_t ** vpd);

typedef device_node_info_t *
(device_get_info_function_t) (const device_t type,
			      const device_bus_info_t bus_info);

typedef char *
(device_get_name_function_t) (const device_t type,
			      const device_bus_info_t bus_info);

typedef void
(device_set_yl_function_t) (const device_t type,
			    const device_bus_info_t bus_info,
			    const char * device_bus_dir,
			    vpd_list_t ** vpd);

typedef bool
(device_add_hook_function_t) (const device_t type,
			      const device_bus_info_t bus_info,
			      const char * device_subtype,
			      char ** device_bus_dir,
			      vpd_list_t ** vpd);

typedef struct device_functions {
	device_t                             type;
	device_t                             supertype;
	device_validate_function_t         * validate;
	device_retrieve_vpd_function_t     * retrieve_vpd;
	device_get_parent_name_function_t  * get_parent_name;
	device_vpd_sort_function_t         * vpd_sort;
	device_get_info_function_t         * get_info;
	device_get_name_function_t         * get_kernel_name;
	device_get_name_function_t         * get_name;
	device_set_yl_function_t           * set_yl;   // FIXME: unused
	device_add_hook_function_t         * add_hook; // FIXME: unused
	struct device_functions            * next;
} device_functions_t;

extern device_functions_t * device_functions;

#define DEVICE_UNDEF ((void *) -1)
#define DEVICE_DEFAULT_TYPE "default"

/*
 * device_type_register:
 *
 * NULL fields in i are ignored.  Fields with value DEVICE_UNDEF cause
 * a previous value of that field for i->type to be NULLed out.  All
 * other non-NULL fields overwrite previous values for i->type.
 */
extern void
device_type_register(device_functions_t * i);

extern device_validate_function_t         device_validate;
extern device_retrieve_vpd_function_t     device_retrieve_vpd;
extern device_get_parent_name_function_t  device_get_parent_name;
extern device_vpd_sort_function_t         device_vpd_sort;
extern device_get_info_function_t         device_get_info;
extern device_get_name_function_t         device_get_name;
extern device_get_name_function_t         device_get_kernel_name;

/********************************************************************/

/* Device listing. */

typedef bool (device_listing_function_t) (device_bus_info_list_t ** device_list);

/* This list is logically unordered. */
typedef struct device_listing_function_list {
	device_listing_function_t * f;
	struct device_listing_function_list * next;
} device_listing_function_list_t;

extern device_listing_function_list_t * device_listing_functions;

extern void
device_listing_functions_clear(void);

extern void
device_listing_function_register(device_listing_function_t * f);

extern bool
device_list_devices(device_bus_info_list_t ** device_list);

extern bool
device_list_add(device_bus_info_list_t ** device_list,
		const device_t type,
		const device_bus_info_t bus_info);

extern void
device_list_clear(device_bus_info_list_t ** device_list);

/********************************************************************/

typedef struct {
	char * name;
	int fd;
} device_temp_t;

extern device_temp_t *
device_open_temporary(const device_node_info_t * info);

extern void
device_close_temporary(device_temp_t ** d);

#endif
