/*
  device_get_vpd - Retrieve VPD from devices.
  
  This file is part of the Linux lsvpd package.
  
  (C) Copyright IBM Corp. 2005
  
  Maintained by  Martin Schwenke <martins@au.ibm.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
 
  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
  $Id: device_handler.c,v 1.7 2005/07/11 00:16:41 martins Exp $
*/

#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <string.h>

#include "init.h"
#include "util.h"
#include "lsvpd.h"
#include "vpd.h"
#include "device.h"

/********************************************************************/

#define DEVICE_LIST              "list_devices"
#define DEVICE_GET_VPD           "get_vpd"
#define DEVICE_PRINT_NAME        "get_name"
#define DEVICE_PRINT_KERNEL_NAME "get_kernel_name"
#define DEVICE_PRINT_PARENT_NAME "get_parent_name"

static void
usage(const char * prog)
{
	fprintf(stderr,
		"usage:"
		"\t%s %s\n"
		"\t%s %s type bus_info tmp_dir vpd_dir\n"
		"\t%s %s type bus_info\n"
		"\t%s %s type bus_info\n"
		"\t%s %s type bus_info\n",
		prog, DEVICE_LIST,
		prog, DEVICE_GET_VPD,
		prog, DEVICE_PRINT_NAME,
		prog, DEVICE_PRINT_KERNEL_NAME,
		prog, DEVICE_PRINT_PARENT_NAME);
	fprintf(stderr, "\nEnvironment variable LSVPD_DB_DIR must be set.\n");
	exit(1);
}


static bool
device_get_vpd(const device_t type,
	       const device_bus_info_t bus_info,
	       const char * tmp_dir,
	       const char * vpd_dir)
{
	bool ret;
	vpd_list_t * vpd;

	ret = false;

	vpd = NULL;

	// FIXME: Note that this creates a redundant .lsvpd file in
	// the VPD directory.  When this module is properly integrated
	// into a larger system, all fields will be collected and
	// sorted before writing.  At the moment, the calling program
	// still adds some VPD and sorts, so we don't bother doing
	// that here.
	ret = device_retrieve_vpd(type, bus_info, tmp_dir, &vpd) &&
		(NULL != vpd) &&
		lsvpd_mkdir(vpd_dir) &&  // FIXME: should be lsvpd_mkdir_safe
		vpd_list_to_vpd_dir(vpd, vpd_dir);

	if (NULL != vpd)
		vpd_list_free(vpd);
	
	return ret;
}

bool device_list(void)
{
	bool ret;
	device_bus_info_list_t * l;
	device_bus_info_list_t * t;

	lsvpd_db_mkdirs();

	l = NULL;
	ret = device_list_devices(&l);

	if (ret)
		for (t = l; NULL != t; t = t->next)
			printf("%s\n", t->bus_info);

	device_list_clear(&l);

	return ret;
}

bool device_print_name(const device_t type,
		       const device_bus_info_t bus_info)
{
	bool ret;
	char * n;

	ret = false;

	n = device_get_name(type, bus_info);
	if (NULL != n) {
		ret = true;
		printf("/dev/%s\n", n);
		free(n);
	}

	return ret;
}

bool device_print_kernel_name(const device_t type,
			      const device_bus_info_t bus_info)
{
	bool ret;
	char * n;

	ret = false;

	n = device_get_kernel_name(type, bus_info);
	if (NULL != n) {
		ret = true;
		printf("/dev/%s\n", n);
		free(n);
	}

	return ret;
}

bool device_print_parent_name(const device_t type,
			      const device_bus_info_t bus_info)
{
	bool ret;
	char * n;

	ret = false;

	n = device_get_parent_name(type, bus_info);
	if (NULL != n) {
		ret = true;
		printf("%s\n", n);
		free(n);
	}

	return ret;
}

int
main(int argc, char ** argv)
{
	bool ret;

	ret = false;

	 /*
	  * Placing the call(s) to call_inits() is an interesting
	  * problem.  If it appears before the following if statement,
	  * then checks in the init functions can stop us making it
	  * through to the usage() message below.  Putting calls to
	  * call_inits() in the above functions means they get buried.
	  * Repeating them here, before each function call makes them
	  * pretty obvious.
	  *
	  * Also, various init() functions dynamically allocate memory
	  * that is not freed.  Since this only happens once, this is
	  * OK.
	  */

	if ((2 == argc) &&
	    (0 == strcmp(argv[1], DEVICE_LIST))) {
		call_inits();
		ret = device_list();
	} else if ((6 == argc) &&
		   (0 == strcmp(argv[1], DEVICE_GET_VPD))) {
		call_inits();
		ret = device_get_vpd(argv[2], argv[3], argv[4], argv[5]);; 
	} else if ((4 == argc) &&
		   (0 == strcmp(argv[1], DEVICE_PRINT_NAME))) {
		call_inits();
		ret = device_print_name(argv[2], argv[3]); 
	} else if ((4 == argc) &&
		   (0 == strcmp(argv[1], DEVICE_PRINT_KERNEL_NAME))) {
		call_inits();
		ret = device_print_kernel_name(argv[2], argv[3]); 
	} else if ((4 == argc) &&
		   (0 == strcmp(argv[1], DEVICE_PRINT_PARENT_NAME))) {
		call_inits();
		ret = device_print_parent_name(argv[2], argv[3]); 
	} else
		usage(argv[0]);

	return ret ? 0 : 1;
}
