/*
  device_ide - VPD support routines for IDE devices.
  
  This file is part of the Linux lsvpd package.
  
  (C) Copyright IBM Corp. 2005
  
  Maintained by  Martin Schwenke <martins@au.ibm.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
 
  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
  $Id: device_ide.c,v 1.3 2005/07/06 09:10:58 martins Exp $
*/

/* For asprintf(3). */
#define _GNU_SOURCE

#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "lsvpd.h"
#include "init.h"
#include "util.h"
#include "device.h"

#include "device_ide.h"

/********************************************************************/

bool
device_parse_ide_address(const char * s,
			 device_ide_address_t * a)
{
	return (2 == sscanf(s, "%u.%u", &a->channel, &a->unit));
}

bool
device_parse_ide_bus_info(const device_bus_info_t bus_info,
			  device_ide_address_t * a)
{
	const char * ba;

	ba = device_find_address(DEVICE_IDE_TYPE, bus_info);
	return ((NULL != ba) && device_parse_ide_address(ba, a));
}

/********************************************************************/

/*
 * Configuration file detailing how TM should be used/modified when
 * determining MF.
 */
#define DEVICE_IDE_MF_MAP "ide_mf.map"

typedef struct {
	char * tm;
	char * mf;

} device_ide_do_mf_tm_t;

static const char * device_ide_tm_skip[] = {
	"CD-ROM",
	NULL,
};

static bool
device_ide_do_mf_tm_scanner(const char * line,
			    void * x)
{
	bool ret;
	const char * s;
	device_ide_do_mf_tm_t * mfl;
	size_t l;
	char * t;
	const char * z;
	int i;

	mfl = x;
	ret = true;

	/* Not a comment, TM matches prefix part of line (up to space). */
	if (('#' != line[0]) &&
	    (NULL != (s = strchr(line, ' '))) &&
	    ('\0' != s[1]) &&
	    (0 < (l = (s - line))) &&
	    (0 == strncasecmp(line, mfl->tm, l))) {

		/* MF is part of line after space.  Finished searching map. */
		mfl->mf = strdup(s + 1);
		ret = false;
		
		/* Skip prefix in TM if followed by ' '. */
		t = mfl->tm; /* Just for brevity. */
		if (' ' == t[l])
			t = t + l + 1;  /* At worst *t is NUL. */

		/* Skip extra spaces. */ 
		while (' ' == *t)
			t++;
		
		/* Anything fixed to skip? */
		i = 0;
		while (NULL != (z = device_ide_tm_skip[i])) {
			l = strlen(z);
			if ((0 == strncasecmp(z, t, l)) &&
			    (' ' == t[l])) {
				t = t + l + 1;
				break;
			}
			i++;
		}

		/* Skip extra spaces. */ 
		while (' ' == *t)
			t++;
		
		/* Skipping anything?  Anything left? */
		if ((t != mfl->tm) && ('\0' != t[0])) {
			l = strlen(t);
			memmove(mfl->tm, t, l);
			mfl->tm[l] = '\0';
		}
	}

	return ret;
}

bool
device_ide_do_mf_tm(char ** mf, char ** tm)
{
	bool ret;
	device_ide_do_mf_tm_t mfl;
	char * mf_map;

	ret = false;

	*mf = NULL;

	mfl.tm = *tm;
	mfl.mf = NULL;
	
	mf_map = lsvpd_relative_filename(lsvpd_lib_dir,
					 DEVICE_IDE_MF_MAP);
	if (NULL != mf_map) {
		if (line_scanner_harness(mf_map,
					 device_ide_do_mf_tm_scanner,
					 &mfl)) {
			/*
			 * proc_ide_do_mf_tm_line_scanner is
			 * responsible for freeing mfl.tm if it
			 * reallocates it.
			 */
			*tm = mfl.tm;
			*mf = mfl.mf;  /* Could still be NULL! */
			ret = (NULL != *mf);
		}
		
		free(mf_map);
	}

	if (NULL == *mf)
		*mf = strdup("UNKNOWN");

	return ret;
}

/********************************************************************/

static void
init(void)
{
	device_functions_t s = 
		{
			.type             = DEVICE_IDE_TYPE,
			.supertype        = DEVICE_DEFAULT_TYPE,
		};
	device_type_register(&s);
}

INIT(init);
