/*
  device - VPD support routines for devices.
  
  This file is part of the Linux lsvpd package.
  
  (C) Copyright IBM Corp. 2005
  
  Maintained by  Martin Schwenke <martins@au.ibm.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
 
  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
  $Id: device_scsi.c,v 1.5 2005/07/09 22:58:39 martins Exp $
*/

/* For asprintf(3). */
#define _GNU_SOURCE

#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "lsvpd.h"
#include "init.h"
#include "util.h"

#include "device.h"

#include "device_scsi.h"

/********************************************************************/

bool
device_parse_scsi_address(const char * s,
			  device_scsi_address_t * a)
{
	return (4 == sscanf(s,
			    "%u:%u:%u:%u",
			    &a->host, &a->bus, &a->target, &a->lun));
}

bool
device_parse_scsi_bus_info(const device_bus_info_t bus_info,
			   device_scsi_address_t * a)
{
	const char * ba;

	ba = device_find_address(DEVICE_SCSI_TYPE, bus_info);
	return ((NULL != ba) && device_parse_scsi_address(ba, a));
}

/********************************************************************/

/* Very vaguely taken from linux-2.5.22/drivers/scsi/scsi.[ch]
 * There are only really 14 possible type, but filling it out to 16
 * removes the need for error checking. */
static const struct string_array device_scsi_types_short[] =
{
        {0x0, "disk"},
        {0x1, "tape"},
        {0x2, "printer"},
        {0x3, "processor"},
        {0x4, "worm"},
        {0x5, "cdrom"},
        {0x6, "scanner"},
        {0x7, "optical-disk"},
        {0x8, "jukebox"},
        {0x9, "communications"},
        {0xa, "graphics"},
        {0xb, "graphics"},
        {0xd, "enclosure"},
        {-1, NULL}
};

static const struct string_array device_scsi_types[] =
{
        {0x0, "Disk Drive"},
        {0x1, "Tape Drive"},
        {0x2, "Printer Device"},
        {0x3, "Processor Device"},
        {0x4, "Write-once Device"},
        {0x5, "CD-ROM Drive"},
        {0x6, "Scanner Device"},
        {0x7, "Optical Memory Device"},
        {0x8, "Medium Changer Device"},
        {0x9, "Communications Device"},
        {0xa, "Graphics Arts Pre-press Device"},
        {0xb, "Graphics Arts Pre-press Device"},
        {0xd, "Enclosure Services Device"},
        {-1, NULL}
};


const char *
device_scsi_subtype_string(unsigned char subtype)
{
	return string_array_lookup(device_scsi_types_short,
				   "unknown",
				   subtype);
}

char *
device_scsi_ds(unsigned char subtype,
	       const char * mf)
{
	char * ret;

	if (-1 == asprintf(&ret,
			   "%s %s",
			   ((NULL != mf) && (0 == strcmp("ATA", mf))) ?
			   "ATA" : "SCSI",
			   /* string_array_lookup can not fail. */
			   string_array_lookup(device_scsi_types,
					       "Unknown Device",
					       subtype)))
		ret = NULL;

	return ret;
}

/********************************************************************/

static char *
device_scsi_hexify(const unsigned char * s,
		   size_t len)
{
	int i;
	char * ret = malloc(len*2+1);
	for (i = 0; i < len ; i++) {
		snprintf(&ret[i*2], 3, "%.2X", s[i]);
	}

	return ret;
}

bool
device_scsi_append_field(vpd_list_t **  vpd,
			 const char *   key,
			 const unsigned char * data,
			 size_t         len,
			 const char *   subtype_string)
{
	bool ret;
	char * ldata;

	ret = false;

	/* Certain fields will require some munging. */
	/* Convert RL, Z0, Z7 to hex. */
	if (((strcmp(key, "RL") == 0) &&
	     (strcmp(subtype_string, "disk") == 0)) ||
	    (strcmp(key, "Z0") == 0) ||
	    (strcmp(key, "Z7") == 0))
		/* device_scsi_hexify returns a new NUL-terminated string. */
		ldata = device_scsi_hexify(data, len);
	else
		/* Make trimmed, NUL-terminated copy. */
		ldata = strndup_trim(data, len);
	
	if (NULL != ldata) {
		/* Write data to file. */
		ret = vpd_list_maybe_append_key_string(vpd, key,
						       ldata);
		free(ldata);
	}

	return ret;
}

bool
device_scsi_append_field_string(vpd_list_t **  vpd,
				const char *   key,
				const char *   data,
				const char *   subtype_string)
{
	return (NULL != data) &&
		device_scsi_append_field(vpd, key,
					 (unsigned char *) data, strlen(data),
					 subtype_string);
}

/********************************************************************/

static bool
device_scsi_validate (const device_t type,
		      const device_bus_info_t bus_info)
{
	bool ret;
	device_scsi_address_t a;

	ret = true;

	if (device_parse_scsi_bus_info(bus_info, &a) &&
	    (0xff == a.bus) &&
	    (0xff == a.target) &&
	    (0xff == a.lun))
		ret = false;

	return ret;
}

static char *
device_scsi_get_parent_name (const device_t type,
			     const device_bus_info_t bus_info)
{
	char * ret;
	device_scsi_address_t a;

	ret = NULL;

	if (device_parse_scsi_bus_info(bus_info, &a)) {
		// FIXME: "host" should not be hardcoded.
		if (-1 == asprintf(&ret, "host%d", a.host))
			ret = NULL;
	}

	return ret;
}

static const char * device_scsi_vpd_sort_order[] =
{
	"FC", "DS", "AX", "PL", "MF", "TM", "CD", "LL",
	"YL", "FN", "RL", "SN", "EC", "PN",
	NULL
};

static void
device_scsi_vpd_sort (const device_t type,
		      vpd_list_t ** vpd)
{
	vpd_list_sort(vpd, device_scsi_vpd_sort_order);
}

/********************************************************************/

/*
 * Defaults:
 *   retrieve_vpd returns false.
 */

static void
init(void)
{
	device_functions_t s = 
		{
			.type             = DEVICE_SCSI_TYPE,
			.supertype        = DEVICE_DEFAULT_TYPE,
			.validate         = device_scsi_validate,
			.get_parent_name  = device_scsi_get_parent_name,
			.vpd_sort         = device_scsi_vpd_sort,
		};
	device_type_register(&s);
}

INIT(init);
