/*
  device - sysfs support routines for devices.
  
  This file is part of the Linux lsvpd package.
  
  (C) Copyright IBM Corp. 2005
  
  Maintained by  Martin Schwenke <martins@au.ibm.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
 
  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
  $Id: device_sysfs.c,v 1.10 2005/07/09 22:53:45 martins Exp $
*/

/* For asprintf(3). */

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <sysfs/libsysfs.h>

#include "util.h"
#include "vpd.h"

#include "lsvpd.h"
#include "init.h"
#include "sysfs.h"

#include "device.h"
#include "device_scsi.h"
#include "device_ide.h"

/********************************************************************/

static struct {
	device_t type;
	char *   sysfs_bus;
} device_sysfs_type_alist[] = {
	{ DEVICE_IDE_TYPE,     "ide"   },
	{ DEVICE_SCSI_TYPE,    "scsi"  },
	{ DEVICE_DEFAULT_TYPE, NULL },
};

static const char *
device_sysfs_type_lookup(const device_t type)
{
	const char * ret;
	int i;

	ret = NULL;

	i = 0;
	while ((NULL == ret) &&
	       (NULL != device_sysfs_type_alist[i].sysfs_bus)) {
		if (0 == strcmp(type, device_sysfs_type_alist[i].type)) {
			ret = device_sysfs_type_alist[i].sysfs_bus;
		}
		i++;
	}

	return ret;
}

/********************************************************************/

static bool
device_sysfs_list_devices(device_bus_info_list_t ** device_list)
{
	bool ret;

	int i;
	device_t t;
	char * s;
	struct dlist * names;
	char * c;
	device_bus_info_t bi;

	ret = true;

	device_list_clear(device_list);

	i = 0;
	while (NULL != (s = device_sysfs_type_alist[i].sysfs_bus)) {
		t = device_sysfs_type_alist[i].type;
		names = sysfs_open_bus_devices_list(s);
		if (NULL != names) {
			dlist_for_each_data(names, c, char) {
				if (-1 != asprintf(&bi, "%s%c%s", t,
						   DEVICE_BUS_INFO_SEP, c)) {
					ret = ret && 
						device_list_add(device_list,
								t, bi);
					free(bi);
				}
			}
		}
		sysfs_close_list(names);
		i++;
	}

	return ret;
}

/********************************************************************/

static struct {
	char * type;
	mode_t mode;
} device_sysfs_links[] = {
	{ "block",   S_IFBLK },
	{ "tape",    S_IFCHR },
	{ "generic", S_IFCHR },
	{ NULL,      S_IFREG },
};

static struct sysfs_link *
device_sysfs_find_link(const device_t type,
		       const device_bus_info_t bus_info)
{
	struct sysfs_link * ret;
	const char * s;
	const char * t;
	struct sysfs_device * d;
	struct sysfs_link * l;
	int i;

	ret = NULL;

	s = device_sysfs_type_lookup(type);
	t = device_find_address(type, bus_info);
	if ((NULL == s) || (NULL == t))
		goto done;

	d = sysfs_open_device(s, t);
	if (NULL == d)
		goto done;

	d->directory = sysfs_open_directory(d->path);
	if (NULL == d->directory)
		goto close_device;

	i = 0;
	l = NULL;
	while ((NULL == l) && (NULL != device_sysfs_links[i].type)) {
		l = sysfs_get_directory_link(d->directory,
					     device_sysfs_links[i].type);
		i++;
	}
					
	if (NULL != l) {
		ret = malloc(sizeof(struct sysfs_link));
		if (NULL != ret)
			memcpy(ret, l, sizeof(struct sysfs_link));
	}

 close_device:
	sysfs_close_device(d);
	
 done:
	return ret;
}

/* Returns S_IFREG on fail. */
static mode_t
device_sysfs_get_mode(const char * type)
{
	mode_t ret;
	int i;

	ret = S_IFREG;

	i = 0;
	while ((S_IFREG == ret) &&
	       (NULL != device_sysfs_links[i].type)) {
		if (0 == strcmp(device_sysfs_links[i].type, type))
			ret = device_sysfs_links[i].mode;
		i++;
	}

	return ret;
}

static struct {
	const char * from;
	const char * to;
} device_sysfs_prefix_subst_alist[] = {
	{ "sr",  "scd" },
	{ NULL,  NULL  },
};

static void
device_sysfs_prefix_substitute(char ** s)
{
	int i;
	const char * f;

	i = 0;
	while ((NULL != (f = device_sysfs_prefix_subst_alist[i].from)) &&
	       !lsvpd_prefix_substitute(s,
					f,
					device_sysfs_prefix_subst_alist[i].to))
		i++;
}

static char *
device_sysfs_get_directory_attribute(const char * d,
				     const char * a)
{
	char * ret;
	struct sysfs_directory * sd;
	char * t;
	struct sysfs_attribute * sa;

	ret = NULL;

	sd = sysfs_open_directory(d);
	if (NULL != sd ) {
		t = strdup(a);
		if (NULL != t) {
			sa = sysfs_get_directory_attribute(sd, t);
			if (NULL != sa)
				ret = strndup_trim((unsigned char *) sa->value,
						   sa->len);
			free(t);
		}
		sysfs_close_directory(sd);
	}

	return ret;
}

/********************************************************************/

static char *
device_sysfs_get_kernel_name(const device_t type,
			     const device_bus_info_t bus_info)
{
	char * ret;
	struct sysfs_link * l;

	ret = NULL;

	l = device_sysfs_find_link(type, bus_info);
	if (NULL != l) {
		ret = lsvpd_basename(l->target);
		device_sysfs_prefix_substitute(&ret);
		free(l);
	}

	return ret;
}

static device_node_info_t *
device_sysfs_get_info(const device_t type,
		      const device_bus_info_t bus_info)
{
	device_node_info_t * ret;
	struct sysfs_link * l;
	char * mm;

	ret = NULL;

	l = device_sysfs_find_link(type, bus_info);
	if (NULL != l) {
		mm = device_sysfs_get_directory_attribute(l->target, "dev");
		if (NULL != mm) {
			ret = malloc(sizeof(device_node_info_t));
			if (NULL != ret) {
				ret->mode = device_sysfs_get_mode(l->name);
				if ((S_IFREG == ret->mode) ||
				    (2 != sscanf(mm, "%u:%u",
						 &ret->major,
						 &ret->minor))) {
					free(ret);
					ret = NULL;
				}
			}
			free(mm);
		}
		free(l);
	}

	return ret;
}

/********************************************************************/

static bool
device_sysfs_retrieve_vpd_scsi(const device_t type,
			       const device_bus_info_t bus_info,
			       const char * tmp_dir,
			       vpd_list_t ** vpd)
{
	bool ret;
	const char * s;
	const char * t;
	struct sysfs_device * d;
	char * st;
	unsigned int x;
	unsigned char subtype;
	const char * subtype_string;
	char * mf;
	char * tm;
	char * rl;
	char * ds;

	ret = false;

	s = device_sysfs_type_lookup(type);
	t = device_find_address(type, bus_info);
	if ((NULL == s) || (NULL == t))
		goto done;

	d = sysfs_open_device(s, t);
	if (NULL == d)
		goto done;

	st = device_sysfs_get_directory_attribute(d->path, "type");
	mf = device_sysfs_get_directory_attribute(d->path, "vendor");
	tm = device_sysfs_get_directory_attribute(d->path, "model");
	rl = device_sysfs_get_directory_attribute(d->path, "rev");

	if (1 != sscanf(st, "%u", &x))
		goto free_fields;
	subtype = (unsigned char) x;
	subtype_string = device_scsi_subtype_string(subtype);

	if (lsvpd_file_relative_write_string(tmp_dir, "type",
					     subtype_string)) {
		ds = device_scsi_ds(subtype, mf);
		if (NULL != ds) {
			ret = true;

			if ((NULL != tm) &&
			    (NULL != mf) &&
			    (0 == strcmp(mf, "ATA"))) {
				free(mf);
				device_ide_do_mf_tm(&mf, &tm);
			}

			device_scsi_append_field_string(vpd, "TM", tm,
							subtype_string);
			device_scsi_append_field_string(vpd, "MF", mf,
							subtype_string);
			device_scsi_append_field_string(vpd, "RL", rl,
							subtype_string);
			device_scsi_append_field_string(vpd, "DS", ds,
							subtype_string);

			free(ds);
		}
	}

 free_fields:
	free(st);
	free(mf);
	free(tm);
	free(rl);

	sysfs_close_device(d);
 done:
	return ret;
}

/********************************************************************/

static void
init(void)
{
	if (NULL != lsvpd_sysfs_dir) {
		device_functions_t d = {
			.type            = DEVICE_DEFAULT_TYPE,
			.get_kernel_name = device_sysfs_get_kernel_name,
			.get_info        = device_sysfs_get_info,
		};

		device_functions_t s = {
			.type            = DEVICE_SCSI_TYPE,
			.get_kernel_name = DEVICE_UNDEF,
			.get_info        = DEVICE_UNDEF,
			.retrieve_vpd    = device_sysfs_retrieve_vpd_scsi,
		};

		device_functions_t i = {
			.type            = DEVICE_IDE_TYPE,
			.get_kernel_name = DEVICE_UNDEF,
			.get_info        = DEVICE_UNDEF,
		};

		device_type_register(&d);
		device_type_register(&s);
		device_type_register(&i);

		device_listing_functions_clear();
		device_listing_function_register(device_sysfs_list_devices);
	}
}

INIT(init);
