/*
  pci_scsi - Get PCI (and other) information for scsi adapters.
  
  This file is part of the Linux lsvpd package.
  
  (C) Copyright IBM Corp. 2004, 2005
  
  Maintained by  Martin Schwenke <martins@au.ibm.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
 
  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
  $Id: pci_scsi.c,v 1.4 2005/07/07 06:45:13 martins Exp $
*/
  
#define _GNU_SOURCE

#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <dirent.h>
#include <unistd.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>

#include "util.h"
#include "pci_scsi.h"

#define PCI_SCSI_ADAPTER_TEMPLATES LSVPD_SUBDIR "/scsihost.conf"
#define PCI_SCSI_PROC_DIR          "/proc/scsi"

typedef struct pci_scsi_io {
	char *         proc_scsi_dir;
	pci_scsi_t *   out;
} pci_scsi_io_t;


void
pci_scsi_fprint(FILE * f, pci_scsi_t * l)
{
	while (NULL != l) {
		fprintf(f, "scsi%d\n", (unsigned int) l->num);
		fprintf(f, "\tbus\t%04lx:%02hhx:%02hhx:%02hhx\n",
			(unsigned long) l->dom, (unsigned char) l->bus,
			(unsigned char) l->dev, (unsigned char) l->fun);
		fprintf(f, "\tirq\t%hu\n", l->irq);

		l = l->next;
	}
}

void
pci_scsi_free(pci_scsi_t * l)
{
	pci_scsi_t * t;
	while (NULL != l) {
		t = l;
		l = l->next;
		free(t);
	}
}

typedef void (*pci_scsi_parser) (uint32_t      hostnum,
				 const char *  driver,
				 const char *  data,
				 pci_scsi_t ** ps);

void pci_scsi_sym538xx_parser (uint32_t      hostnum,
			       const char *  driver,
			       const char *  data,
			       pci_scsi_t ** ps)
{
	unsigned int  bus;
	unsigned char dev;
	unsigned char fun;
	unsigned int  irq;

	pci_scsi_t * p = NULL;
	char * t;

	if (NULL == (t = strstr(data, "PCI bus")))
		return;
	if (4 != sscanf(t, "PCI bus %u, device %hhu, function %hhu, IRQ %u",
			&bus, &dev, &fun, &irq))
		return;

	p = calloc(1, sizeof(pci_scsi_t));

	p->num = hostnum;

	if (bus > 0xff) {
		p->bus = (uint8_t) (bus & 0xff);
		p->dom = bus >> 8;
	} else {
		p->bus = (uint8_t) bus;
		p->dom = PCI_SCSI_NODOM;
	}
	p->dev = dev;
	p->fun = fun;

	p->irq = irq;
	p->driver = driver;

	p->next = *ps;
	*ps = p;
}

void pci_scsi_ipr_parser (uint32_t      hostnum,
			  const char *  driver,
			  const char *  data,
			  pci_scsi_t ** ps)
{
	unsigned long long  hostaddr;

	pci_scsi_t * p = NULL;
	char * t;

	if (NULL == (t = strstr(data, "Host Address:")))
		return;
	if (1 != sscanf(t, "Host Address: %llx", &hostaddr))
		return;

	p = calloc(1, sizeof(pci_scsi_t));

	p->num = hostnum;

	p->fun = (uint8_t) (hostaddr & 0xff);
	p->dev = (uint8_t) ((hostaddr >> 8) & 0xff);
	p->bus = (uint8_t) ((hostaddr >> 16) & 0xff);
	p->dom = (uint32_t) ((hostaddr >> 24) & 0xffffff);

	/* Doesn't appear to have leading 0s, so we just don't know! :-( */
	if (0 == p->dom)
		p->dom = PCI_SCSI_NODOM;

	p->irq = 0;
	p->driver = driver;

	p->next = *ps;
	*ps = p;
}

struct pci_scsi_driver_info {
	char * name;
	pci_scsi_parser parser;
};

static struct pci_scsi_driver_info pci_scsi_drivers[] = {
	{ "sym53c8xx", pci_scsi_sym538xx_parser },
	{ "ipr",       pci_scsi_ipr_parser },
	{ "ibmsis",    pci_scsi_ipr_parser },
	{ NULL,        NULL }
};

/* This is horrible!  If any of the entries for a particular driver
 * have a known PCI domain, then we assume the driver prints long bus
 * numbers, so we set any unknown domains to 0.  This can save the
 * caller a lot of trouble. */
static void
hack_doms(const char * driver, pci_scsi_t * l)
{
	pci_scsi_t * t;
	bool hackdoms = false;

	for (t = l; NULL != t; t = t->next)
		if ((0 == strcmp(t->driver, driver)) &&
		    (PCI_SCSI_NODOM != t->dom)) {
			hackdoms = true;
			break;
		}
	if (hackdoms)
		for (t = l; NULL != t; t = t->next)
			if ((0 == strcmp(t->driver, driver)) &&
			    (PCI_SCSI_NODOM == t->dom))
				t->dom = 0;
}

/* Extra level of functions to allow testing against fixed files. */
static pci_scsi_t *
pci_scsi_file(const char * proc_scsi_dir)
{
	struct pci_scsi_driver_info * p = pci_scsi_drivers;
	char * dname = NULL;
	DIR * dir;
	struct dirent * e;
	unsigned int hostnum;
	char x;
	char * f = NULL;
	char * data = NULL;
	
	pci_scsi_t * ret = NULL;

	p = pci_scsi_drivers;

	for (p = pci_scsi_drivers; NULL != p->name; p++) {
		/*fprintf(stderr, "%s\n", p->name);*/

		/* dname = /proc/scsi/<driver>. */
		asprintf(&dname, "%s/%s", proc_scsi_dir, p->name);

		dir = opendir(dname);
		if (NULL == dir) {
			free(dname);
			continue;
		}

		while (NULL != (e = readdir(dir))) {
			if (1 != sscanf(e->d_name, "%u%c", &hostnum, &x))
				continue;
			asprintf(&f, "%s/%u", dname, hostnum);
			data = lsvpd_file_read_string(f);
			if (NULL != data) {
				p->parser((uint32_t) hostnum, p->name,
					  data, &ret);
				free(data);
			}
			free(f);
		}

		free(dname);
		closedir(dir);
	}

	for (p = pci_scsi_drivers; NULL != p->name; p++)
		hack_doms(p->name, ret);

	return ret;
}


pci_scsi_t *
pci_scsi()
{
	return pci_scsi_file(PCI_SCSI_PROC_DIR);
}

#if 0
int
main(int argc, char ** argv)
{
	if (2 != argc) {
		fprintf(stderr, "usage: pci_scsi file\n");
		return 1;
	}

	pci_scsi_t * l = pci_scsi_file(argv[1]);

	pci_scsi_fprint(stderr, l);
	pci_scsi_free(l);

        return 0;
}
#endif
