/*
  pci_scsi_map - Print PCI (and other) information for scsi adapters.
  
  This file is part of the Linux lsvpd package.
  
  (C) Copyright IBM Corp. 2004
  
  Maintained by  Martin Schwenke <martins@au.ibm.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
 
  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
  $Id: pci_scsi_map.c,v 1.5 2004/10/19 05:56:25 martins Exp $
*/

#include <stdbool.h>
#include <stdio.h>
#include <linux/pci_ids.h>
  
#include "pci_scsi.h"
#include "proc_pci.h"

static void
print_item(uint32_t num, uint32_t dom,  uint8_t bus, uint8_t dev, uint8_t fun)
{
	printf("%04x:%02hhx:%02hhx.%hhx host%u\n",
	       (unsigned int)  dom,
	       (unsigned char) bus,
	       (unsigned char) dev,
	       (unsigned char) fun,
	       (unsigned int)  num);
}

static bool
is_possible_match(pci_scsi_t * s, proc_pci_list_t * p)
{
	return ((p->class == PCI_CLASS_STORAGE_SCSI) &&
		(s->bus == (p->bus & 0xff))          &&
		(s->dev == p->dev)                   &&
		(s->fun == p->fun)                   &&
		((0 == s->irq) || (s->irq == (p->irq & 0xff))));
}

static bool
pci_unique_scsi_match(proc_pci_list_t * p, pci_scsi_t * sl)
{
	pci_scsi_t * t;
	unsigned int num_matches = 0;

	for (t = sl; NULL != t; t = t->next)
		if (is_possible_match(t, p)) {
			num_matches++;
			if (1 < num_matches)
				return false;
		}

	return (1 == num_matches);
}

static bool
scsi_unique_pci_match(pci_scsi_t * s, proc_pci_list_t * pl)
{
	proc_pci_list_t * t;
	unsigned int num_matches = 0;

	for (t = pl; NULL != t; t = t->next)
		if (is_possible_match(s, t)) {
			num_matches++;
			if (1 < num_matches)
				return false;
		}

	return (1 == num_matches);
}

static void
print_dodgy_matches(pci_scsi_t * s, pci_scsi_t * sl)
{
	pci_scsi_t * st;
	proc_pci_list_t * t;
	uint32_t dom;
	proc_pci_list_t * pl = proc_pci();

	for (t = pl; NULL != t; t = t->next)
		if (is_possible_match(s, t)) {
			dom = (uint32_t) (t->bus >> 8) & 0xffffff;
			if (pci_unique_scsi_match(t, sl))
				/* PCI address matches one SCSI host. */
				print_item(s->num,
					   dom, s->bus, s->dev, s->fun);
			else
				/* Find a different SCSI host that
				   matches the address we're looking
				   at.  If this other SCSI host
				   matches only that address, then the
				   address can't belong to the current
				   adapter - otherwise print it.  */
				for (st = sl; NULL != st; st = st->next)
					if ((s->num != st->num) &&
					    is_possible_match(st, t) &&
					    (!scsi_unique_pci_match(st, pl)))
						print_item(s->num,
							   dom, s->bus,
							   s->dev, s->fun);
		}

	proc_pci_list_free(pl);
}

int
main(int argc, char ** argv)
{
	pci_scsi_t  * sl;
	pci_scsi_t  * t;

	if (NULL == (sl = pci_scsi()))
		return 1;

	for (t = sl; NULL != t; t = t->next) {
		if (PCI_SCSI_NODOM != t->dom)
			print_item(t->num, t->dom, t->bus, t->dev, t->fun);
		else {
			print_dodgy_matches(t, sl);
		}
	}

	pci_scsi_free(sl);

        return 0;
}
