/*
  pci_vpd_cap_grab.c - Grab PCI VPD via the capabilities list.
  
  This file is part of the Linux lsvpd package.
  
  (C) Copyright IBM Corp. 2002, 2003, 2004, 2005
  
  Maintained by  Martin Schwenke <martins@au.ibm.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
 
  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
  $Id: pci_vpd_cap_grab.c,v 1.15 2005/05/24 04:56:33 martins Exp $
*/

/* For pread, pwrite: */
#define _XOPEN_SOURCE 500

#include <sys/stat.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <netinet/in.h>
#include <linux/pci.h>

#include "growbuf.h"
#include "util.h"
#include "pci_vpd_gen_grab.h"

static void
usage(const char * prog)
{
	fprintf(stderr, "usage: %s file\n", prog);
	exit(1);
}

/********************************************************************/

/* These should never give short reads or writes */
ssize_t
big_cool_pread(int fd, uint8_t *buf, size_t count, off_t offset)
{
	ssize_t total = 0;
	ssize_t last  = 0;

	do {
		last = pread(fd, buf + total, count - total, offset + total);
		if (-1 != last) {
			total += last;
		}
	} while ((0 < last) && (total < count));
	
	if ((-1 == last) && (0 == total)) {
		total = -1;
	}

	return total;
}

ssize_t
big_cool_pwrite(int fd, const uint8_t *buf, size_t count, off_t offset)
{
	ssize_t total = 0;
	ssize_t last  = 0;

	do {
		last = pwrite(fd, buf + total, count - total, offset + total);
		if (-1 != last) {
			total += last;
		}
	} while ((0 < last) && (total < count));
	
	if ((-1 == last) && (0 == total)) {
		total = -1;
	}

	return total;
}

/********************************************************************/

struct vpd_cap_state{
	struct vpd_state s;
	int              fd;
	uint8_t          cap_ptr;
	uint16_t         vpd_offset;
	uint8_t          buf[4];
};

static struct vpd_cap_state *
vpd_cap_create(vpd_read_func rf, int fd, uint8_t cap_ptr)
{
	struct vpd_cap_state * ret = malloc(sizeof(struct vpd_cap_state));
	
	if (NULL != ret) {
		ret->s.rf       = rf;
		ret->fd         = fd;
		ret->cap_ptr    = cap_ptr;
		ret->vpd_offset = 0;
	}

	return ret;
}

static void
vpd_cap_destroy(struct vpd_cap_state * s)
{
	free(s);
}

static ssize_t
vpd_cap_read(struct vpd_state * this, uint8_t * c_ptr, ssize_t count)
{
	struct vpd_cap_state * x = (struct vpd_cap_state * ) this;

	ssize_t i = 0;

	while (i < count) {

		/* Bail out if top bit is set.  This bit is used as a
		 * flag. */
		if (0 != (x->vpd_offset & (1 << 15))) {
			fprintf(stderr,
				"vpd_cap_read: warning - "
				"VPD offset value exceeds 15 bits=%02x\n",
				x->vpd_offset);
			return -1;
		}

		/* x->vpd_offset % 4 always points to the next byte
		 * in the 4 byte window, so if that is 0 we have no
		 * data ready and we need to fetch some.
		 */
		if (0 == (x->vpd_offset % 4)) {

			short n = 0;
			ssize_t c;
			
			/* 1. Write the VPD Address at offset 2. */
			uint16_t a = cpu_to_le16(x->vpd_offset);
			if (2 != big_cool_pwrite(x->fd,
						 (uint8_t *) &a,
						 2, x->cap_ptr + 2))
				return -1;
			
			/* 2. Poll until the flag is set. */
			while ((2 == (c = big_cool_pread(x->fd,
							 (uint8_t *) &a,
							 2,
							 x->cap_ptr + 2))) &&
			       (0 == (le16_to_cpu(a) & (1 << 15)))) {
				/* Just spin, but not for too long. */
				 if (n == 500)
					 sleep(1);
				 if (n == 1000)
					 return -1;
				 n++;
			}

			if (2 != c)
				return -1;
			
			/* 3. Read 4 bytes of VPD Data. */
			if (4 != big_cool_pread(x->fd, x->buf,
						4, x->cap_ptr + 4))
				return -1;
		}
		
		/* Grab next available byte. */
		c_ptr[i] = x->buf[x->vpd_offset % 4];
		
		x->vpd_offset++;
		i++;
	}
		
	return i;
}

static bool
have_capabilities(int fd)
{
	uint16_t status;
	bool ret = false;

	if (-1 != big_cool_pread(fd, (uint8_t *) &status, 2, PCI_STATUS)) {
		ret = (le16_to_cpu(status) & PCI_STATUS_CAP_LIST) != 0;
	}

	return ret;
}


/* Return an "array" of offsets to VPD capability structures.  It
 * doesn't say anywhere that a device (well, function) can have
 * multiple VPD structures, but it also doesn't say that you can't...
 * The logic is no more complex, so let's just do it.
 */

static growbuf *
find_vpd_capabilities(int fd)
{
	uint8_t buf[2];
	uint8_t type, pos;

	growbuf * ret = growbuf_new(16);

	/* Determine the position of the capability list. */
	if (1 == big_cool_pread(fd, &type, 1, PCI_HEADER_TYPE)) {
		type &= ~(1 << 7);
		switch (type) {
		case PCI_HEADER_TYPE_NORMAL:
		case PCI_HEADER_TYPE_BRIDGE:
			pos = PCI_CAPABILITY_LIST;
			break;
		case PCI_HEADER_TYPE_CARDBUS:
			pos = PCI_CB_CAPABILITY_LIST;
			break;
		default:
			fprintf(stderr,
				"find_vpd_capabilities: warning - "
				"unknown PCI header type=%02x\n",
				type);
			pos = 0;
		}
	} else {
		pos = 0;
	}

	if ((0 != pos) && (1 == big_cool_pread(fd, &pos, 1, pos))) {
		/* Traverse list, remembering positions of any VPD
		 * capability structures.
		 */
		while ((0 != pos) &&
		       (2 == big_cool_pread(fd, buf, 2, pos))) {

			if (PCI_CAP_ID_VPD == buf[0]) {
				ret = growbuf_append(ret, &pos, 1);
			}
			pos = buf[1];
		}
	}

	return ret;
}

int
main(int argc, char ** argv)
{
	int fd;
	int ret = 0;
	growbuf * cap_ptrs;

	growbuf * vpd = growbuf_new(1024);

	if (2 != argc) {
		usage(argv[0]);
	}

	fd = open(argv[1], O_RDWR);
	if (-1 == fd) {
		fprintf(stderr, "%s: %s (%s)\n", argv[0],
			strerror(errno), argv[1]);
		exit(1);
	}

	if ((have_capabilities(fd)) &&
	    (NULL != (cap_ptrs = find_vpd_capabilities(fd))) &&
	    (0 < growbuf_length(cap_ptrs))) {

		growbuf_iter * i = growbuf_iter_new(cap_ptrs);
		while ((0 == ret) && (growbuf_iter_more(i))) {
			
			struct vpd_cap_state * s =
				vpd_cap_create(vpd_cap_read,
					       fd,
					       *growbuf_iter_current(i));
			growbuf * frag = grab_vpd((struct vpd_state *) s);
			vpd_cap_destroy(s);

			if (NULL != frag) {
				/* Push length onto VPD, as in ibm,vpd. */
				uint32_t length = htonl(growbuf_length(frag));
				vpd = growbuf_append(vpd,
						     (uint8_t *)&length,
						     sizeof(uint32_t));
				/* Append frag to VPD. */
				vpd = growbuf_append(vpd,
						     growbuf_data(frag),
						     growbuf_length(frag));
				
				growbuf_free(frag);
			} else {
				ret = 1;
			}

			growbuf_iter_next(i);
		}

		growbuf_iter_free(i);
		growbuf_free(cap_ptrs);
		
	} else {
		ret = 1;
	}

	close(fd);

	if (NULL != vpd) {
		fwrite(growbuf_data(vpd), growbuf_length(vpd), 1, stdout);
		growbuf_free(vpd);
	} else {
		ret = 1;
	}

	return ret;
}
