#ifndef VPD_UTIL_H
#define VPD_UTIL_H

/*
  util - VPD utilitiy functions.
  
  This file is part of the Linux lsvpd package.
  
  (C) Copyright IBM Corp. 2002, 2003, 2004, 2005
  
  Maintained by  Martin Schwenke <martins@au.ibm.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
 
  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
  $Id: util.h,v 1.28 2005/07/09 23:01:13 martins Exp $
*/
  
#include <stdio.h>
#include <stdint.h>
#include <stdbool.h>
#include <sys/types.h>

/********************************************************************/

/* Filename manipulation. */

/* Construct filename dir/relpath.  Must free(3) non-NULL result. */
extern char *
lsvpd_relative_filename(const char * dir,
			const char * relpath);

/* Like basename(3), dirname(3) but doesn't modify arg, and creates a copy. */
extern char *
lsvpd_basename(const char * f);
extern char *
lsvpd_dirname(const char * f);

/*
 * Like realpath(3) but only works if path is a directory.  Uses
 * malloc(3) to allocate space for result.  path must exist and must
 * not contain any broken symlinks, since chdir(2) is use in this
 * function.
 */
char *
lsvpd_realpath(const char * path);

char *
lsvpd_realpath_relative(const char * dir,
			const char * path);
/********************************************************************/

/* Directory creation and removal. */

/* Like mkdir(1) -p. */
extern bool
lsvpd_mkdir(const char * dir);

/* lsvpd_mkdir(dirname(f)); */
extern bool
lsvpd_mkdirname(const char * f);

/*
 * Safely remove a directory tree.  A sanity check ensures that
 * argument prefix is actually a prefix of d.
 */
extern void
lsvpd_rm_rf(const char * prefix,
	    const char * d);

/********************************************************************/

/* File I/O. */

extern bool
file_read_contents(const char *     pathname,
		   unsigned char ** buf,
		   size_t *         length,
		   int  *           err);

extern bool
file_relative_read_contents(const char *     dir,
			    const char *     relpath,
			    unsigned char ** buf,
			    size_t *         length,
			    int *            err) ;

extern char *
lsvpd_file_read_string(const char * pathname);

extern char *
lsvpd_file_relative_read_string(const char * dir,
				const char * relpath);
extern bool
file_write_contents(const char  *         pathname,
		    const unsigned char * buf,
		    size_t                count,
		    int *                 err);

extern bool
file_relative_write_contents(const char *          dir,
			     const char *          relpath,
			     const unsigned char * data,
			     size_t                count,
			     int *                 err);

extern bool
lsvpd_file_write_string(const char * pathname,
			const char * string);

extern bool
lsvpd_file_relative_write_string(const char * dir,
				 const char * relpath,
				 const char * string);

/********************************************************************/

/* Some of the shifts and masks below are redundant, but that's why we
 * have an optimiser!  Isn't it nice?  :-)
 */

static inline uint16_t
le16_to_cpu(uint16_t n)
{
	return 	((uint16_t)(((uint8_t*)&n)[1]) << 8) |
		((uint16_t)(((uint8_t*)&n)[0]) << 0) ;
}

static inline uint16_t
cpu_to_le16(uint16_t n)
{
	uint16_t ret;
	uint8_t * t = (uint8_t *) &ret;

	t[0] = (n >> 0) & 0xff;
	t[1] = (n >> 8) & 0xff;

	return ret;
}

static inline uint32_t
le32_to_cpu(uint32_t n)
{
	return 	((uint32_t)(((uint8_t*)&n)[3]) << 24) |
		((uint32_t)(((uint8_t*)&n)[2]) << 16) |
		((uint32_t)(((uint8_t*)&n)[1]) <<  8) |
		((uint32_t)(((uint8_t*)&n)[0]) <<  0) ;
}

static inline uint32_t
cpu_to_le32(uint32_t n)
{
	uint32_t ret;
	uint8_t * t = (uint8_t *) &ret;

	t[0] = (n >>  0) & 0xff;
	t[1] = (n >>  8) & 0xff;
	t[2] = (n >> 16) & 0xff;
	t[3] = (n >> 24) & 0xff;

	return ret;
}

/********************************************************************/

/* link readlink(3), but malloc(3) result. */
extern char *
lsvpd_readlink(const char * l);

/* string_array is useful for making sparse, fixed arrays without
 * needing to do bounds checking. */

struct string_array {
	int key;
	const char * val;
};

extern const char *
string_array_lookup(const struct string_array * a,
		    const char *                d,
		    int                         k);

extern char *
strndup_trim(const unsigned char * s,
	     size_t n);

/* If s starts with from, replace from with to. */
bool
lsvpd_prefix_substitute(char **      s,
			const char * from,
			const char * to);
extern bool
line_scanner_harness(const char * filename,
		     bool (* line_scanner) (const char *, void *),
		     void *       result);

#endif
